module SQLParser where
import qualified SQLModel as SQL
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Expr
import qualified Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Language( emptyDef )
import List (nub, intersperse)

intercalate x xs = concat $ intersperse x xs
-----------------------------------------------------------
-- The Lexer
-----------------------------------------------------------
lexer     = P.makeTokenParser sqlDef

indenters = [["SELECT"], ["FROM"], ["WHERE"], ["ORDER", "BY"], ["GROUP", "BY"]]


reservedOpNames =  [".", "?", "=", "<>", "<", "<=", ">=", ">", "+", "-", "/", "\\", "*",
                                "#", "#>", "#<", "#<"]

-- , P.opLetter       = oneOf $ nub $ concat $ reservedOpNames
-- , P.opStart        = oneOf $ nub $ concat $ (map (take 1) reservedOpNames)
sqlDef  = emptyDef
        { P.commentStart = "/*"
        , P.commentEnd   = "*/"
        , P.commentLine  = "--"     
        , P.nestedComments = False
        , P.identStart   = alphaNum <|> oneOf "_.*"     
        , P.identLetter  = alphaNum <|> oneOf "_.*"     
                  
        , P.reservedNames  = nub $ concat $ indenters
        , P.reservedOpNames= reservedOpNames
		, P.opLetter       = oneOf "="
        , P.caseSensitive  = False
        }

lexeme          = P.lexeme lexer
whiteSpace      = P.whiteSpace lexer
parens          = P.parens lexer    
braces          = P.braces lexer    
colon           = P.colon  lexer
commaSep1       = P.commaSep1 lexer    
symbol          = P.symbol lexer    
identifier      = P.identifier lexer    
reserved        = P.reserved lexer    
reservedOp      = P.reservedOp lexer
operator        = P.operator lexer
integer         = P.integer lexer    
comma           = P.comma lexer

seqReserved xs = foldl1 (>>) (map reserved xs)

res :: [String] -> CharParser st String
res xs  = do 
	seqReserved xs
	return $ intercalate " " xs

oneOfParsers parser xs =  foldl1 (<|>) (map parser xs)

indenterWords = oneOfParsers res indenters

reservedOp' x = do 
	reservedOp x
	return x 

reservedOps = oneOfParsers reservedOp' reservedOpNames
                   
indenter = makeItem SQL.Indenter indenterWords

par = makeItem SQL.Parens (parens sqlSequence)

makeItem constructor strParser  = do
	x <- strParser
	return $ constructor x

makeText x = makeItem SQL.Text x

iden = makeText  identifier

comma_p = makeText comma

reservedOp_p = makeText reservedOps
-- reservedOp_p = reservedOps  

{-- >> -- ப     '  ୠ '' ''  '      --}
str1chars = do {a <- noneOf "'"; return (a:"") } <|> try (string "''")
strLiteral1 = lexeme (  
    do{ str <- between (char '\'')
        (char '\'' <?> "end of literal_1 string")
        (many str1chars)
        ; return ("'"++(foldr (++) "" str)++"'")
    }
    <?> "literal_1 string")

{-- >> -- ப     "  ୠ "" ""  "      --}
str2chars = do {a <- noneOf "\""; return (a:"") } <|> try (string "\"\"")
strLiteral2 = lexeme (  
    do{ str <- between (char '"')
        (char '"' <?> "end of literal_2 string")
        (many str2chars)
        ; return ("\""++(foldr (++) "" str)++"\"")
    }
    <?> "literal_2 string")


stringLiteral = strLiteral1 <|> strLiteral2
stringLit = makeText stringLiteral


someItem = reservedOp_p <|> indenter <|> par <|> comma_p <|> stringLit <|> iden

sqlSequence = many1 someItem 


sqlParser = do 
	whiteSpace
	x <- sqlSequence
	eof
	return x